@@ -46,6 +46,7 @@ struct RBGICallbackData_ {
     RBGICallback *callback;
     RBGIArgMetadata *metadata;
     VALUE rb_callback;
+    GObject *owner;
 };
 
 static VALUE RG_TARGET_NAMESPACE;
@@ -303,18 +304,6 @@ fill_metadata(GPtrArray *args_metadata, GICallableInfo *info)
     fill_metadata_rb_arg_index(args_metadata);
 }
 
-static void
-callback_data_guard_from_gc(RBGICallbackData *callback_data)
-{
-    rbg_gc_guard(callback_data, callback_data->rb_callback);
-}
-
-static void
-callback_data_unguard_from_gc(RBGICallbackData *callback_data)
-{
-    rbg_gc_unguard(callback_data);
-}
-
 static void
 rb_gi_callback_free(RBGICallback *callback)
 {
@@ -325,13 +314,30 @@ rb_gi_callback_free(RBGICallback *callback)
     xfree(callback);
 }
 
+static void
+rb_gi_callback_data_weak_notify(gpointer data, GObject* where_the_object_was)
+{
+    RBGICallbackData *callback_data = data;
+    callback_data->owner = NULL;
+}
+
 void
 rb_gi_callback_data_free(RBGICallbackData *callback_data)
 {
     if (callback_data->callback) {
         rb_gi_callback_free(callback_data->callback);
     }
-    callback_data_unguard_from_gc(callback_data);
+    if (callback_data->owner) {
+        VALUE rb_owner;
+        g_object_weak_unref(callback_data->owner,
+                            rb_gi_callback_data_weak_notify,
+                            callback_data);
+        rb_owner = rbgobj_ruby_object_from_instance2(callback_data->owner,
+                                                     FALSE);
+        if (!NIL_P(rb_owner)) {
+            rbgobj_object_remove_relative(rb_owner, callback_data->rb_callback);
+        }
+    }
     xfree(callback_data->metadata);
     xfree(callback_data);
 }
@@ -349,7 +355,7 @@ rb_gi_callback_data_get_rb_callback(RBGICallbackData *callback_data)
 }
 
 static void
-destroy_notify(gpointer data)
+rb_gi_callback_data_destroy_notify(gpointer data)
 {
     RBGICallbackData *callback_data = data;
     rb_gi_callback_data_free(callback_data);
@@ -1087,7 +1093,10 @@ ffi_closure_callback(G_GNUC_UNUSED ffi_cif *cif,
 }
 
 static void
-in_callback_argument_from_ruby(RBGIArgMetadata *metadata, GArray *in_args)
+in_callback_argument_from_ruby(RBGIArgMetadata *metadata,
+                               GArray *in_args,
+                               VALUE self,
+                               VALUE rb_arguments)
 {
     gpointer callback_function;
     GIArgInfo *arg_info;
@@ -1140,17 +1149,39 @@ in_callback_argument_from_ruby(RBGIArgMetadata *metadata, GArray *in_args)
 
     if (closure_argument) {
         RBGICallbackData *callback_data;
+        VALUE rb_owner;
 
         callback_data = ALLOC(RBGICallbackData);
         callback_data->callback = callback;
         callback_data->metadata = metadata;
         callback_data->rb_callback = rb_block_proc();
-        callback_data_guard_from_gc(callback_data);
-        closure_argument->v_pointer = callback_data;
+        if (NIL_P(self)) {
+            rb_owner = rb_ary_entry(rb_arguments, 0);
+        } else {
+            rb_owner = self;
+        }
+        {
+            static VALUE mGLibObject = Qnil;
+            if (NIL_P(mGLibObject)) {
+                mGLibObject = rb_const_get(mGLib, rb_intern("Object"));
+            }
+            if (rb_obj_is_kind_of(rb_owner, mGLibObject)) {
+                rbgobj_object_add_relative(rb_owner, callback_data->rb_callback);
+                callback_data->owner = RVAL2GOBJ(rb_owner);
+                g_object_weak_ref(callback_data->owner,
+                                  rb_gi_callback_data_weak_notify,
+                                  callback_data);
+            } else {
+                /* Class method case. Callback is never GC-ed. (OK?) */
+                rbgobj_add_relative(rb_owner, callback_data->rb_callback);
+                callback_data->owner = NULL;
+            }
+            closure_argument->v_pointer = callback_data;
+        }
     }
 
     if (destroy_argument) {
-        destroy_argument->v_pointer = destroy_notify;
+        destroy_argument->v_pointer = rb_gi_callback_data_destroy_notify;
     }
 }
 
@@ -1163,7 +1194,7 @@ in_argument_from_ruby(GICallableInfo *callable_info,
                       VALUE self)
 {
     if (metadata->callback_p && !metadata->destroy_p) {
-        in_callback_argument_from_ruby(metadata, in_args);
+        in_callback_argument_from_ruby(metadata, in_args, self, rb_arguments);
         return;
     }
 
